#include <list>
#include <iterator>
#include "ActionScheduler.h"

using namespace std;

/**
 * Basic constructor of action scheduler
 */
ActionScheduler::ActionScheduler(WorldModel* wm, ServerSettings *ss, PlayerSettings *ps)
{
	WM         = wm;
	SS         = ss;
	PS         = ps;

#ifdef DEBUG
	LOG_PRINT(LOG_INFO, "Initializing ActionScheduler");
#endif //DEBUG
}

/**
 * Performs cleanup of action queue. All parameters that are asociated with action function
 * are cleaned up using cleanup handler
 */
ActionScheduler::~ActionScheduler()
{
#ifdef DEBUG
	LOG_PRINT(LOG_INFO, "Calling ActionScheduler destructor");
#endif //DEBUG

	while( ! scheduled_actions.empty() ) {
		struct scheduled_action *action = &(scheduled_actions.front());
		
		if(action->cleanup_callback != NULL)
			(action->cleanup_callback)(&(action->params), action->params_size);
		
		scheduled_actions.pop();
	}
}

/**
 * Executes all actions, that are asociated with cycle. Actions are cleaned up from queue. If
 * call is performed again, actions wont be executed.
 */
void ActionScheduler::dispatchAction(int cycle) {
#ifdef DEBUG
	LOG_PRINTP(LOG_DEBUG, "Executing actions - cycle %d", cycle);
#endif //DEBUG

	while( ! scheduled_actions.empty() ) {
		struct scheduled_action *action = &(scheduled_actions.front());
		
		if(action->cycle > cycle) 
			break;

#ifdef DEBUG		
		LOG_PRINTP(LOG_DEBUG, "Dispatching action - method address: %p, cleanup_callback: %p",
			action->action_callback, action->cleanup_callback );
#endif //DEBUG
			
		doDispatchAction(action);
		
		scheduled_actions.pop();
	}
}

/**
 * Internal action dispatching mechanism
 */
void ActionScheduler::doDispatchAction(struct scheduled_action *action) {
	(action->action_callback) (WM, SS, PS, action->params, action->params_size);
	if( action->cleanup_callback != NULL )
		(action->cleanup_callback) (action->params, action->params_size);
}

/**
 * Scheduling of action.
 * @param cycle Identifies cycle to schedule action execution to. It is absolute number. 
 * @param action_callback Method that contains actual code responsible for action evaluation
 * @param cleanup_callback Method, that should be used for action parameters cleanup. If NULL passed
 * 		no parameters cleanup is performed
 * @param params Parameters that are to be passed to action_callback
 * @param params_size Size of segment that parameters attribute is pointing to.
 */
void ActionScheduler::scheduleAction(
		int cycle, 
		void (*action_callback)(WorldModel*,ServerSettings*,PlayerSettings*, void* params, int params_size), 
		void (*cleanup_callback)(void*,int), 
		void* params,
		int params_size) {
	
#ifdef DEBUG
	LOG_PRINTP(LOG_DEBUG, "Scheduling action_callback: %p, cleanup_callback: %p, for cycle %d",
		action_callback, cleanup_callback, cycle );
#endif //DEBUG
	
	struct scheduled_action action;
	action.cycle = cycle;
	action.action_callback = action_callback;
	action.cleanup_callback = cleanup_callback;
	action.params = params;
	action.params_size = params_size;
	
	scheduled_actions.push(action);
}
