#include <math.h>
#include "ArsToolbox.h"

/** GOS - Determines cross point (intersection) of two lines
 * 
 * Determines intersection of two lines by using one
 * parameterized formula and nominating it's X and Y
 * to formula of second line
 * 
 * @return 0 if 2 points have been found, 1 if no solution, 2 if circle are identical
 * 
 * @param x1 X coordinate of start of 1st line
 * @param y1 Y coordinate of start of 1st line
 * @param x2 X coordinate of start of 2nd line
 * @param y2 Y coordinate of start of 2nd line
 * @param x3 X coordinate of end of 1st line 
 * @param y3 Y coordinate of end of 1st line
 * @param x4 X coordinate of end of 2nd line
 * @param y4 Y coordinate of end of 2nd line
 * @param cpx intersection X
 * @param cpy intersection Y
 * 
 */
int ArsToolbox::lineLineCrossPoint(double x1, double y1, double x2, double y2, double x3, double y3,
					  double x4, double y4, double *cpx, double *cpy)
{
	double ua, ub;
	double men;
	
	ua =  ((x4-x3)*(y1-y3) - (y4-y3)*(x1-x3));
	ub =  ((x2-x1)*(y1-y3) - (y2-y1)*(x1-x3));
	men = ((y4-y3)*(x2-x1) - (x4-x3)*(y2-y1));
	
	// usecky su paralelne
	if(men == 0)
	{
		return 1;
	}
	// usecky su zhodne
	else if(ua == 0 && ub == 0 && men == 0)
	{
		return 2;
	}
	// crosspoint existuje a jeho suradnice su (cpx, cpy)
	else
	{
		ua /= men;
		ub /= men;
		*cpx = (x1 + ua*(x2-x1));
		*cpy = (y1 + ub*(y2-y1));
	}
	
	return 0;
}

/** GOS - Determines intersection between two circles
 * 
 * Determines intersection of two circles by using one
 * parameterized circle formula and nominating it's X and Y
 * to formula of second circle
 * 
 * @return 0 if 2 points have been found, 1 if no solution, 2 if circle are identical
 *
 * @param x0 X coordinate of centre of 1st circle
 * @param y0 Y coordinate of centre of 1st circle
 * @param r0 radius of 1st circle
 * @param x1 X coordinate of centre of 2nd circle
 * @param y1 Y coordinate of centre of 2nd circle
 * @param r1 radius of 2nd circle
 * @param xi X of first intersection
 * @param yi Y of first intersection
 * @param xi_prime X of second intersection
 * @param yi_prime Y of secound intersection
 *  
 */
int ArsToolbox::circleCircleIntersection(double x0, double y0, double r0,
                               double x1, double y1, double r1,
                               double *xi, double *yi,
                               double *xi_prime, double *yi_prime)
{
  double a, dx, dy, d, h, rx, ry;
  double x2, y2;

  /* dx and dy are the vertical and horizontal distances between
   * the circle centers.
   */
  dx = x1 - x0;
  dy = y1 - y0;

  /* Determine the straight-line distance between the centers. */
  d = sqrt((dy*dy) + (dx*dx));

  /* Check for solvability. */
  if (d > (r0 + r1))
  {
    /* no solution. circles do not intersect. */
    return 1;
  }
  if (d < fabs(r0 - r1))
  {
    /* no solution. one circle is contained in the other */
    return 2;
  }

  /* 'point 2' is the point where the line through the circle
   * intersection points crosses the line between the circle
   * centers.  
   */

  /* Determine the distance from point 0 to point 2. */
  a = ((r0*r0) - (r1*r1) + (d*d)) / (2.0 * d) ;

  /* Determine the coordinates of point 2. */
  x2 = x0 + (dx * a/d);
  y2 = y0 + (dy * a/d);

  /* Determine the distance from point 2 to either of the
   * intersection points.
   */
  h = sqrt((r0*r0) - (a*a));

  /* Now determine the offsets of the intersection points from
   * point 2.
   */
  rx = -dy * (h/d);
  ry = dx * (h/d);

  /* Determine the absolute intersection points. */
  *xi = x2 + rx;
  *xi_prime = x2 - rx;
  *yi = y2 + ry;
  *yi_prime = y2 - ry;

  return 0;
}

/** GOS - Returns angle of two vectors
 * 
 * Determines angle A between two vectors V1, V2:
 * cos(A) = (V1.x*V2.x + V1.y*V2.y) / (vectorNorm(V1) * vectorNorm(V2)) 
 *
 * @return angle of two vectors in radians(!)
 *  
 * @param v1ex X of V1
 * @param v1ey Y of V1
 * @param v2ex X of V2
 * @param v2ey Y of V2
 * 
 */
double ArsToolbox::vectorsAngle(double v1ex, double v1ey, double v2ex, double v2ey)
{
	return acos((v1ex*v2ex + v1ey*v2ey) / (sqrt((v1ex*v1ex) + (v1ey*v1ey))*sqrt((v2ex*v2ex) + (v2ey*v2ey))));	
}

/** GOS - Returns vector size
 * 
 * Returns vector size:
 * size = sqrt(x^2 + y^2)
 * 
 * @return vector size(length)
 * 
 * @param v1x x coordinate
 * @param v1y y coordinate
 */
double ArsToolbox::vectorNorm(double v1x, double v1y)
{
	return sqrt((v1x*v1x) + (v1y*v1y));
}

/** Determines if line and circle have crossing points
 * 
 * This function creates line formula as follows:
 * x = p1x + (p2x - p1x)*t
 * y = p1y + (p2y - p1y)*t
 * 
 * And then places it to circle formula:
 * (x - cx)^2 + (y - cy)^2 = r^2
 * 
 * Solution is kvadratic function with no, 1 or 2 solutions (points)
 * 
 * @return number of solutions
 * @param line if true, then points [p1x, p1y], [p2x, p2y] are for line, segment (dash) otherwise  
 * 
 */
int ArsToolbox::circleLineIntersection(double p1x, double p1y, double p2x, double p2y, double cx, double cy, double r, double *v1x, double *v1y, double *v2x, double *v2y)
{
    ArsToolbox::circleLineIntersection(p1x, p1y, p2x, p2y, cx, cy, r, v1x, v1y, v2x, v2y, false);
}

/** GOS - Determines if line and circle have crossing points
 * 
 * This function creates line formula as follows:
 * x = p1x + (p2x - p1x)*t
 * y = p1y + (p2y - p1y)*t
 * 
 * And then places it to circle formula:
 * (x - cx)^2 + (y - cy)^2 = r^2
 * 
 * Solution is kvadratic function with no, 1 or 2 solutions (points)
 * 
 * @return number of solutions
 * @param line if true, then points [p1x, p1y], [p2x, p2y] are for line, segment (dash) otherwise  
 * 
 */
int ArsToolbox::circleLineIntersection(double p1x, double p1y, double p2x, double p2y, double cx, double cy, double r, double *v1x, double *v1y, double *v2x, double *v2y, bool line)
{
    if (p1x < p2x)
    {
        // switch x
        double tmp = p1x;
        p1x = p2x;
        p2x = tmp;
        // switch y
        tmp = p1y;
        p1y = p2y;
        p2y = tmp;
    }
    
    // x1:x2, y1:y2
    double dx = p2x - p1x;
    double dy = p2y - p1y;

    // circle
    double A = dx * dx + dy * dy;
    double B = 2 * (dx * (p1x - cx) + dy * (p1y - cy));
    double C = (p1x - cx) * (p1x - cx) + (p1y - cy) * (p1y - cy) - r * r;

    // determinant of matrix 2x2
    double det = B * B - 4 * A * C;

    // zero points    
    if ((A <= 0.0000001) || (det < 0))
        return 0;
    // exactly 1 point
    else if (det == 0)
    {
        double t = -B / (2 * A);
        if ((t >= 0.0 && t <= 1.0 /*&& p1x <= p2x*/) ||
//            (t <= 0.0 && t >= -1.0 && p1x >= p2x) ||
            (line))
        {
            *v1x = p1x + t * dx;
            *v1y = p1y + t * dy;
            *v2x = (*v1x);
            *v2y = (*v1y);
            return 1;
        }
        else
            return 0;
    }
    // two points
    else
    {
        int count = 0;
        double t;
        t = (-B + sqrt(det)) / (2 * A);
        if ((t >= 0.0 && t <= 1.0 /*&& p1x <= p2x*/) ||
//            (t <= 0.0 && t >= -1.0 && p1x >= p2x) ||
            (line))
        {
            *v1x = p1x + t * dx;
            *v1y = p1y + t * dy;
            count++;
        }
        t = (-B - sqrt(det)) / (2 * A);
        if ((t >= 0.0 && t <= 1.0 /*&& p1x <= p2x*/) ||
//            (t <= 0.0 && t >= -1.0 && p1x >= p2x) ||
            (line))
        {
            if (count == 1)
            {
                *v2x = p1x + t * dx;
                *v2y = p1y + t * dy;
            }
            else
            {
                *v1x = p1x + t * dx;
                *v1y = p1y + t * dy;
            }
            count++;
        }
        return count;
    }        
}

/** GOS - Creates end point for line (segment) from point, angle and length
 * 
 * Function determines vector from angle and length, then 
 * adds original position to this vector
 * 
 * @param sx original X
 * @param sy original Y
 * @param angle angle of vector
 * @param length length of vector
 * @param ex pointer to endpoint X variable 
 * @param eY pointer to endpoint Y variable
 *  
 */
int ArsToolbox::createLineTo(double sx, double sy, double angle, double length, double *ex, double *ey)
{
    *ex = cos(angle) * length + sx;
    *ey = sin(angle) * length + sy;    
}

double ArsToolbox::linePointDistance(double dX1, double dY1, double dX2, double dY2,
    double dPX, double dPY)
{
    // paralell with X,Y coordinates is simple solution
    if (dX1 == dX2)
        return fabs(dPX - dX1);
    if (dY1 == dY2)
        return fabs(dPY - dY1);
        
    // y = Ax + C; Ax + By + C = 0 => A = ?, B = -1, C = 0 
    double A = (dY2 - dY1) / (dX2 - dX1);
    double C = dY1 - dX1 * A;
    
    return fabs(A*dPX - dPY + C) / sqrt(A*A + 1);
} 

double ArsToolbox::trim(double value, double min, double max)
{
    if (value < min)
        return min;
    else if (value > max)
        return max;
    else
        return value;
}

/** GOS - Method computes the distance from a point to a line
	 * 
	 * @param point The point the distance is computets from.
	 * @param lineStart The point where the line starts.
	 * @param lineEnd The point where the line ends.
	 * @param *distance The pointer where to store the distance.
	 * 
	 * @return 0 if no error occures. In case of error 1 is returned.
	 */

int ArsToolbox::distancePointLine(VecPosition point, VecPosition lineStart, VecPosition lineEnd, double * distance) 
{
	double lineMag;
    double U;
    VecPosition intersection;
 
    lineMag =(lineEnd - lineStart).getMagnitude();
 
    U = ( ( ( point.getX() - lineStart.getX() ) * ( lineEnd.getX() - lineStart.getX() ) ) +
        ( ( point.getY() - lineStart.getY() ) * ( lineEnd.getY() - lineStart.getY() ) ) ) /
        ( lineMag * lineMag );
 
    if( U < 0.0 || U > 1.0 )
        return 1;   // closest point does not fall within the line segment
 
    intersection.setX(lineStart.getX() + U * ( lineEnd.getX() - lineStart.getX() ));
    intersection.setY(lineStart.getY() + U * ( lineEnd.getY() - lineStart.getY() ));
    
    *distance = (point-intersection).getMagnitude();
 
    return 0;
}

/** GOS - Method counts time (using Newton's iteration method) player needs to get to the ball.
		*
		*	@param b0 Actual ball position
		*	@param v0 Actual velocity vector
		*   @param r0 Actual player position
		*   @param vR Maximum possible player speed (parameter of soccer server)
		*   @param tau Value used in expression to count dumped velocity of the ball. Implicit value is set to 2.0
		* 
		*   @return Time the player on the position <b>r0</b> needs to get to the ball.
		*/

double ArsToolbox::computeLeastInterceptTime(VecPosition b0, VecPosition v0, VecPosition r0, double vR, double tau) 
{
	VecPosition pTold;
	VecPosition U;
	
	double tOld, tNew;
	double sNew;
	double d;
	double gDerivT;
	double gTold;
	
	int check = 0;
	
	if(v0.getMagnitude() == 0) {
		d = (r0 - b0).getMagnitude();	
	}
	else
	{
		VecPosition v0Norm = v0;
		v0Norm.normalize();
	
		check = ArsToolbox::distancePointLine(r0, b0 + (v0Norm * (-300)), (b0 + (v0Norm * 300)), &d);
	}
	
	// closest point does not fall within the line segment
	if(check == 1) {
		return -1.0;
	}
	
	// t0
	tOld = d / vR;
	
	do {
		// zaciatok cyklu
		pTold = (b0 + ((v0 * tau) * (1 - exp((-tOld)/tau)))); 
		gTold = ((pTold - r0).getMagnitude() - (vR * tOld));

		U = (pTold - r0);
		U.normalize();
		

		// skalarny sucin
		VecPosition ss = ((v0 * exp(-tOld / tau)) * U);
		double skalarnySucin = (ss.getX() + ss.getY());
		
		gDerivT = (skalarnySucin - vR);
		sNew = tOld + (gDerivT / (vR - skalarnySucin));

		// podmienka
		ss = (v0 * U);		
		double skalarnySucinPodmienka = (ss.getX() + ss.getY());
		
		if(skalarnySucinPodmienka < 0) {
			tNew = tOld + (gTold / (vR - skalarnySucin));
		}
		else if(skalarnySucinPodmienka > 0 && gDerivT < 0) {
			VecPosition pSNew = (b0 + ((v0 * tau) * (1 - exp((-sNew)/tau)))); 
			
			VecPosition UNew = pSNew - r0;
			UNew.normalize();

			ss = ((v0 * exp(-tOld / tau)) * UNew);
			double skalarnySucinSNew = (ss.getX() + ss.getY());
			
			tNew = tOld + (gTold / (vR - skalarnySucinSNew));
		}
		else {
			tNew = tOld + gTold / vR;
		}		
		tOld = tNew;
		
	} while(gTold > 0.1);
	
	return tOld;	
}
