/*
 Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.

 3. Neither the name of the University of Amsterdam nor the names of its
 contributors may be used to endorse or promote products derived from this
 software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*! \file Player.cpp

 <pre>
 <b>File:</b>          Player.cpp
 <b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
 <b>Authors:</b>       Jelle Kok
 <b>Created:</b>       03/03/2001
 <b>Last Revision:</b> $ID$
 <b>Contents:</b>      This file contains the definitions for the Player class,
 which is a superclass from BasicPlayer and contains the
 decision procedure to select the skills from the
 BasicPlayer.
 <hr size=2>
 <h2><b>Changes</b></h2>
 <b>Date</b>             <b>Author</b>          <b>Comment</b>
 03/03/2001       Jelle Kok       Initial version created
 </pre>
 */
#include "Player.h"
#include "Parse.h"

#ifndef WIN32
#include <sys/poll.h> // needed for 'poll'
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h> 

extern Logger LogDraw;

/*!This is the constructor the Player class and calls the constructor of the
 superclass BasicPlayer.
 \param act ActHandler to which the actions can be sent
 \param wm WorldModel which information is used to determine action
 \param ss ServerSettings that contain parameters used by the server
 \param ps PlayerSettings that contain parameters important for the client
 \param strTeamName team name of this player
 \param dVersion version this basicplayer corresponds to
 \param iReconnect integer that defines player number (-1 when new player) */
Player::Player(ActHandler* act, WorldModel *wm, ServerSettings *ss,
		PlayerSettings *ps, Formations *fs, char* strTeamName, double dVersion,
		int iReconnect)

{
	log = new SituationsLog(fs->getPlayerInFormation(), strTeamName);

	char str[MAX_MSG];

	ACT = act;
	WM = wm;
	SS = ss;
	PS = ps;
	formations = fs;
	bContLoop = true;
	m_iPenaltyNr = -1;
	WM->setTeamName(strTeamName);
	m_timeLastSay = -5;
	m_objMarkOpp = OBJECT_ILLEGAL;
	m_actionPrev = ACT_ILLEGAL;

	skuska = 1;

	goalMiddle.setVecPosition(PITCH_LENGTH/2, 0, CARTESIAN);

	// wait longer as role number increases, to make sure players appear at the
	// field in the correct order
#ifdef WIN32
	Sleep(formations->getPlayerInFormation()*500);
#else
	poll( 0, 0, formations->getPlayerInFormation()*500 );
#endif

	// create initialisation string
	if (iReconnect != -1)
		sprintf(str, "(reconnect %s %d)", strTeamName, iReconnect);
	else if (formations->getPlayerType() == PT_GOALKEEPER)
		sprintf(str, "(init %s (version %f) (goalie))", strTeamName, dVersion);
	else
		sprintf(str, "(init %s (version %f))", strTeamName, dVersion);
	ACT->sendMessage(str);

}

Player::~Player() {
	delete log;
}

/*! This is the main loop of the agent. This method calls the update methods
 of the world model after it is indicated that new information has arrived.
 After this, the correct main loop of the player type is called, which
 puts the best soccer command in the queue of the ActHandler. */
void Player::mainLoop() {
	char str[MAX_MSG];
	Timing timer;

	// wait for new information from the server
	// cannot say bContLoop=WM->wait... since bContLoop can be changed elsewhere
	if (WM->waitForNewInformation() == false)
		bContLoop = false;

	// and set the clang version
	sprintf(str, "(clang (ver 8 8))");
	ACT->sendMessage(str);

	while (bContLoop) // as long as server alive
	{
		Log.logWithTime( 3, "  start update_all");
		//    Log.setHeader( WM->getCurrentCycle(), WM->getPlayerNumber() );
		Log.setHeader(WM->getCurrentCycle() );
		LogDraw.setHeader(WM->getCurrentCycle() );

		if (WM->updateAll() == true) {
			timer.restartTime();
			SoccerCommand soc;

			if ( (WM->isPenaltyUs() || WM->isPenaltyThem() ))
				performPenalty();
			else if (WM->getPlayMode() == PM_FROZEN)
				ACT->putCommandInQueue(turnBodyToObject(OBJECT_BALL) );
			else {
				switch (formations->getPlayerType() ) // determine right loop
				{
				case PT_GOALKEEPER:
					soc = goalieMainLoop();
					break;
				case PT_DEFENDER_SWEEPER:
				case PT_DEFENDER_CENTRAL:
				case PT_DEFENDER_WING:
					soc = defenderMainLoop();
					break;
				case PT_MIDFIELDER_CENTER:
				case PT_MIDFIELDER_WING:
					soc = midfielderMainLoop();
					break;
				case PT_ATTACKER:
				case PT_ATTACKER_WING:
					soc = attackerMainLoop();
					break;
				case PT_ILLEGAL:
				default:
					break;
				}

				if (shallISaySomething(soc) == true) // shall I communicate
				{
					m_timeLastSay = WM->getCurrentTime();
					char strMsg[MAX_SAY_MSG];
					if (WM->getPlayerNumber() == 6 && WM->getBallPos().getX() < - PENALTY_X + 4.0)
						sayOppAttackerStatus(strMsg);
					else
						sayBallStatus(strMsg);
					if (strlen(strMsg) != 0)
						Log.log( 600, "send communication string: %s", strMsg);
					WM->setCommunicationString(strMsg);
				}
			}
			Log.logWithTime( 3, "  determined action; waiting for new info");
			// directly after see message, will nog get better info, so send commands
			if (WM->getTimeLastSeeMessage() == WM->getCurrentTime()
					|| (SS->getSynchMode() == true && WM->getRecvThink()
							== true )) {
				Log.logWithTime( 3, "  send messages directly");
				ACT->sendCommands();
				Log.logWithTime( 3, "  sent messages directly");
				if (SS->getSynchMode() == true) {
					WM->processRecvThink( false);
					ACT->sendMessageDirect("(done)");
				}
			}

		} else
			Log.logWithTime( 3,
					"  HOLE no action determined; waiting for new info");

		if (WM->getCurrentCycle()%(SS->getHalfTime()*SS->getSimulatorStep())
				!= 0) {
			if (LogDraw.isInLogLevel( 600) ) {
				WM->logDrawInfo( 600);
			}

			if (LogDraw.isInLogLevel( 601) )
				WM->logDrawBallInfo( 601);

			if (LogDraw.isInLogLevel( 700) )
				WM->logCoordInfo( 700);
		}

		Log.logWithTime( 604, "time for action: %f", timer.getElapsedTime()
				*1000);

		// wait for new information from the server cannot say
		// bContLoop=WM->wait... since bContLoop can be changed elsewhere
		if (WM->waitForNewInformation() == false)
			bContLoop = false;
	}

	// shutdow, print hole and number of players seen statistics
	printf("Shutting down player %d\n", WM->getPlayerNumber() );
	printf("   Number of holes: %d (%f)\n", WM->iNrHoles, ((double)WM->iNrHoles
			/WM->getCurrentCycle())*100);
	printf("   Teammates seen: %d (%f)\n", WM->iNrTeammatesSeen,
			((double)WM->iNrTeammatesSeen/WM->getCurrentCycle()));
	printf("   Opponents seen: %d (%f)\n", WM->iNrOpponentsSeen,
			((double)WM->iNrOpponentsSeen/WM->getCurrentCycle()));

}

/*! This is the main decision loop for the goalkeeper. */
SoccerCommand Player::goalieMainLoop() {
	return deMeer5_goalie();
}

/*! This is the main decision loop for a defender. */
SoccerCommand Player::defenderMainLoop() {
	return deMeer5() ;
}

/*! This is the main decision loop for a midfielder. */
SoccerCommand Player::midfielderMainLoop() {
	return deMeer5() ;
}

/*! This is the main decision loop for an agent. */
SoccerCommand Player::attackerMainLoop() {
	return deMeer5() ;
}

/*! This method returns the position to move in case of a dead ball situation.
 A dead ball situation occurs when the team can have a free kick, kick in,
 etc. The agent will move to the position behind the ball and when he is
 there will move to the ball again. */
VecPosition Player::getDeadBallPosition() {
	VecPosition pos, posBall = WM->getBallPos();
	VecPosition posAgent = WM->getAgentGlobalPosition();
	double dDist;

	// determine point to move to
	if (WM->isKickInUs() )
		pos = posBall + VecPosition( -1.5, sign(posBall.getY() )*1.5);
	else if (WM->isCornerKickUs() )
		pos = posBall + VecPosition( 1.5, sign(posBall.getY() ) * 1.5);
	else if (WM->isFreeKickUs() || WM->isOffsideThem() || WM->isGoalKickUs()
			|| WM->isFreeKickFaultThem() || WM->isBackPassThem() )
		pos = posBall + VecPosition( -1.5, 0.0);
	else
		return VecPosition(UnknownDoubleValue, UnknownDoubleValue);

	AngDeg angBall = (posBall-posAgent).getDirection() ;
	ObjectT obj = WM->getClosestInSetTo(OBJECT_SET_PLAYERS,
			WM->getAgentObjectType(), &dDist);
	VecPosition posPlayer = WM->getGlobalPosition(obj);

	// change point when heading towards other player or towards the ball
	if (fabs(angBall - (posPlayer-posAgent).getDirection() ) < 20 && dDist < 6)
		pos -= VecPosition( 5, 0);
	if (fabs(angBall - (pos-posAgent).getDirection()) < 20) {
		angBall = VecPosition::normalizeAngle(angBall - 90);
		pos = posBall + VecPosition( 1, angBall, POLAR);
	}
	return pos;
}

/*!This method listens for input from the keyboard and when it receives this
 input it converts this input to the associated action. See
 showStringCommands for the possible options. This method is used together
 with the SenseHandler class that sends an alarm to indicate that a new
 command can be sent. This conflicts with the method in this method that
 listens for the user input (fgets) on Linux systems (on Solaris this isn't a
 problem). The only known method is to use the flag SA_RESTART with this
 alarm function, but that does not seem to work under Linux. If each time
 the alarm is sent, this gets function unblocks, it will cause major
 performance problems. This function should not be called when a whole match
 is played! */
void Player::handleStdin() {
	char buf[MAX_MSG];

	while (bContLoop) {
#ifdef WIN32
		cin.getline(buf, MAX_MSG);
#else
		fgets( buf, MAX_MSG, stdin ); // does unblock with signal !!!!!
#endif
		printf("after fgets: %s\n", buf);
		executeStringCommand(buf);
	}
}

/*!This method prints the possible commands that can be entered by the user.
 The whole name can be entered to perform the corresponding command, but
 normally only the first character is sufficient. This is indicated by
 putting brackets around the part of the command that is not needed.
 \param out output stream to which the possible commands are printed */
void Player::showStringCommands(ostream& out) {
	out << "Basic commands:" << endl << " a(ctions)" << endl
			<< " c(atch) direction" << endl << " cs(lientsettings" << endl
			<< " d(ash) power [ times ]" << endl << " de(bug) nr_cycles"
			<< endl << " g(oto) x y" << endl << " h(elp)" << endl
			<< " i(ntercept) x y" << endl << " k(ick) power angle" << endl
			<< " ka x y endspeed " << endl << " m(ove) x y" << endl
			<< " n(eck) angle" << endl << " o(pponents in cone) width dist"
			<< endl << " p(redict cycles to) x y" << endl << " q(uit)" << endl
			<< " s(ay) message" << endl << " ss(erversettings)" << endl
			<< " t(urn) angle" << endl
			<< " v(iewmode) narrow | normal | wide low | high" << endl
			<< " w(orldmodel)" << endl;
}

/*!This method executes the command that is entered by the user. For the
 possible command look at the method showStringCommands.
 \param str string that is entered by the user
 \return true when command could be executed, false otherwise */
bool Player::executeStringCommand(char *str) {
	SoccerCommand socCommand;
	int i;
	double x, y;

	switch (str[0]) {
	case 'a': // actions
		WM->showQueuedCommands();
		break;
	case 'c': // catch dir or cs
		if (strlen(str) > 1 && str[1] == 's') {
			PS->show(cout, ":");
			break;
		}
		socCommand.makeCommand(CMD_CATCH, Parse::parseFirstInt( &str) );
		break;
	case 'd': // dash
		socCommand.commandType = CMD_DASH;
		socCommand.dPower = Parse::parseFirstDouble( &str);
		socCommand.iTimes = Parse::parseFirstInt( &str);
		if (socCommand.iTimes == 0)
			socCommand.iTimes = 1;
		break;
	case 'h': // help
		showStringCommands(cout);
		return true;
	case 'k': // kick or ka (kick advanced)
		socCommand.commandType = CMD_KICK;
		if (str[1] == 'a') // advanced kick
		{
			double x = Parse::parseFirstDouble( &str);
			double y = Parse::parseFirstDouble( &str);
			double e = Parse::parseFirstDouble( &str);
			socCommand = kickTo(VecPosition(x, y), e);
		} else {
			socCommand.dPower = Parse::parseFirstDouble( &str);
			socCommand.dAngle = Parse::parseFirstDouble( &str);
		}
		break;
	case 'm': // move
		socCommand.commandType = CMD_MOVE;
		socCommand.dX = Parse::parseFirstDouble( &str);
		socCommand.dY = Parse::parseFirstDouble( &str);
		socCommand.dAngle = Parse::parseFirstDouble( &str);
		break;
	case 'n': // turn_neck
		socCommand.commandType = CMD_TURNNECK;
		socCommand.dAngle = Parse::parseFirstDouble( &str);
		break;
	case 'o': // count nr opp in cone
		x = Parse::parseFirstDouble( &str);
		y = Parse::parseFirstDouble( &str);
		i
				= WM->getNrInSetInCone(OBJECT_SET_OPPONENTS, x,
						WM->getAgentGlobalPosition(),
						WM->getAgentGlobalPosition() + VecPosition(y,
								WM->getAgentGlobalNeckAngle(), POLAR) );
		printf("%d opponents\n", i);
		return true;
	case 'p': // predict cycles to point
		x = Parse::parseFirstDouble( &str);
		y = Parse::parseFirstDouble( &str);
		i = WM->predictNrCyclesToPoint(WM->getAgentObjectType(), VecPosition(x,
				y) );
		printf("%d cycles\n", i);
		return true;
	case 'q': // quit
		bContLoop = false;
		return true;
	case 's': // ss (serversettings) or say
		if (strlen(str) > 1 && str[1] == 's') {
			SS->show(cout, ":");
			break;
		}
		socCommand.commandType = CMD_SAY;
		Parse::gotoFirstOccurenceOf( ' ', &str);
		Parse::gotoFirstNonSpace( &str);
		strcpy(socCommand.str, str);
		break;
	case 't': // turn
		socCommand.commandType = CMD_TURN;
		socCommand.dAngle = Parse::parseFirstDouble( &str);
		break;
	case 'v': // change_view
		socCommand.commandType = CMD_CHANGEVIEW;
		Parse::gotoFirstOccurenceOf(' ', &str);
		Parse::gotoFirstNonSpace( &str);
		socCommand.va = SoccerTypes::getViewAngleFromStr(str);
		Parse::gotoFirstOccurenceOf(' ', &str);
		Parse::gotoFirstNonSpace( &str);
		socCommand.vq = SoccerTypes::getViewQualityFromStr(str);
		break;
	case 'w': // worldmodel
		WM->show();
		return true;
	default: // default: send entered string
		ACT->sendMessage(str);
		return true;
	}
	if (socCommand.commandType != CMD_ILLEGAL) // when socCommand is set
		ACT->putCommandInQueue(socCommand); // send it.

	return true;
}

/*!This method can be called in a separate thread (see pthread_create) since
 it returns a void pointer. It is assumed that this function receives a
 BasicPlayer class as argument. The only thing this function does
 is starting the method handleStdin() from the corresponding BasicPlayer
 class that listens to user input from the keyboard. This function is
 necessary since a method from a class cannot be an argument of
 pthread_create.
 \param v void pointer to a BasicPlayer class.
 \return should never return since function handleStdin has infinite loop*/
#ifdef WIN32
DWORD WINAPI stdin_callback(LPVOID v)
#else
void* stdin_callback( void * v )
#endif

{
	Log.log( 1, "Starting to listen for user input");
	Player* p = (Player*)v;
	p->handleStdin();
	return 0;
}

/********************** SAY **************************************************/

/*!This method determines whether a player should say something.
 \return bool indicating whether the agent should say a message */
bool Player::shallISaySomething(SoccerCommand socPri) {
	bool bReturn;

	bReturn = ((WM->getCurrentTime() - m_timeLastSay) >= SS->getHearDecay());
	bReturn &= amIAgentToSaySomething(socPri);
	bReturn &= (WM->getCurrentCycle() > 0 );

	return bReturn;
}

/*! This method returns a boolean indicating whether I should communicate my
 world model to the other agents.
 \return boolean indicating whether I should communicate my world model. */
bool Player::amIAgentToSaySomething(SoccerCommand socPri) {
	double dDist;
	ObjectT obj;

	// get the closest teammate to the ball, if we are not him, we do not
	// communicate since he will
	obj = WM->getClosestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL, &dDist);
	if (dDist < SS->getVisibleDistance() && obj != WM->getAgentObjectType() )
		return false;

	// in the defense, player 6 keeps track of the opponent attacker
	if (WM->getBallPos().getX() < - PENALTY_X + 4.0 && WM->getConfidence(OBJECT_BALL)
			> 0.96 && WM->getPlayerNumber() == 6 && WM->getCurrentCycle() % 3
			== 0) // once very 3 cycles is enough
	{
		Log.log( 600, "player 6 is going to communicate attacker info");
		return true;
	}

	VecPosition posBallPred;
	WM->predictBallInfoAfterCommand(socPri, &posBallPred);
	VecPosition posAgentPred = WM->predictAgentPosAfterCommand(socPri);
	// in all other cases inform teammates of ball when you have good information
	if ( (WM->getTimeChangeInformation(OBJECT_BALL) == WM->getCurrentTime()
			&& WM->getRelativeDistance(OBJECT_BALL) < 20.0
			&& WM->getTimeLastSeen(OBJECT_BALL) == WM->getCurrentTime() )
			|| (WM->getRelativeDistance(OBJECT_BALL) < SS->getVisibleDistance()
					&& WM->getTimeLastSeen(OBJECT_BALL) == WM->getCurrentTime() )
			|| ( // pass ball
			WM->getRelativeDistance(OBJECT_BALL) < SS->getMaximalKickDist()
					&& posBallPred.getDistanceTo(posAgentPred)
							> SS->getMaximalKickDist() ))
		return true;

	return false;
}

/*! This method encodes the opponent attacker status in a visual message.
 \return string in which the opponent attacker position is encoded.*/
void Player::sayOppAttackerStatus(char* strMsg) {
	char strTmp[MAX_SAY_MSG];

	// fill the first byte with some encoding to indicate the current cycle.
	// The second byte   is the last digit of the cycle added to 'a'.
	sprintf(strMsg, "%c", 'a' + WM->getCurrentCycle()%10);

	// fill the second byte with information about the offside line.
	// Enter either a value between a-z or A-Z indicating. This gives 52 possible
	// values which correspond with meters on the field. So B means the offside
	// line lies at 27.0.
	int iOffside = (int)WM->getOffsideX();
	if (iOffside < 26) // 0..25
		sprintf(strTmp, "%c", 'a' + iOffside);
	else
		// 26...
		sprintf(strTmp, "%c", 'A' + max(iOffside - 26, 25));
	strcat(strMsg, strTmp);

	// find the closest opponent attacker to the penalty spot.
	double dDist;
	ObjectT objOpp = WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, VecPosition(
			- PITCH_LENGTH/2.0 + 11.0, 0), &dDist) ;

	if (objOpp == OBJECT_ILLEGAL || dDist >= 20.0) {
		strncpy(strMsg, "", 0);
		return;
	}

	VecPosition posOpp = WM->getGlobalPosition(objOpp);
	if (fabs(posOpp.getY() ) > 10) {
		strncpy(strMsg, "", 0);
		return;
	}

	// encode the position of this attacker in the visual message. The
	// player_number is the third byte, then comes the x position in 3 digits (it
	// is assumed this value is always negative), a space and finally the y
	// position in 2 digits. An example of opponent nr. 9 at position
	// (-40.3223,-3.332) is "j403 -33)
	sprintf(strTmp, "%c%d %d", 'a' + SoccerTypes::getIndex(objOpp) ,
			(int)(fabs(posOpp.getX())*10), (int)(posOpp.getY()*10));
	strcat(strMsg, strTmp);

	return;
}

/*! This method creates a string to communicate the ball status. When
 the player just kicks the ball, it is the new velocity of the
 ball, otherwise it is the current velocity.
 \param strMsg will be filled */
void Player::sayBallStatus(char * strMsg) {
	VecPosition posBall = WM->getGlobalPosition(OBJECT_BALL);
	VecPosition velBall = WM->getGlobalVelocity(OBJECT_BALL);
	int iDiff = 0;
	SoccerCommand soc = ACT->getPrimaryCommand();

	if (WM->getRelativeDistance(OBJECT_BALL) < SS->getMaximalKickDist() ) {
		// if kick and a pass
		if (soc.commandType == CMD_KICK) {
			WM->predictBallInfoAfterCommand(soc, &posBall, &velBall);
			VecPosition posAgent = WM->predictAgentPos( 1, 0);
			if (posBall.getDistanceTo(posAgent) > SS->getMaximalKickDist()
					+ 0.2)
				iDiff = 1;
		}

		if (iDiff == 0) {
			posBall = WM->getGlobalPosition(OBJECT_BALL);
			velBall.setVecPosition( 0, 0);
		}
	}
	Log.log( 600,
			"create comm. ball after: (%1.2f,%1.2f)(%1.2f,%1.2f) diff %d",
			posBall.getX(), posBall.getY(), velBall.getX(), velBall.getY(),
			iDiff);
	makeBallInfo(posBall, velBall, iDiff, strMsg);
}

/*! This method is used to create the communicate message for the
 status of the ball, that is its position and velocity is encoded.

 \param VecPosition posBall ball position
 \param VecPosition velBall ball velocity
 \param iDiff time difference corresponding to given ball information
 \strMsg string message in which the ball information is encoded. */
void Player::makeBallInfo(VecPosition posBall, VecPosition velBall, int iDiff,
		char * strMsg) {
	char strTmp[MAX_SAY_MSG];

	// fill the first byte with some encoding to indicate the next cycle.
	// The second byte is the last digit of the cycle added to 'a'.
	sprintf(strMsg, "%c", 'a' + (WM->getCurrentCycle()+iDiff)%10);

	// fill the second byte with information about the offside line.
	// Enter either a value between a-z or A-Z indicating. This gives 52 possible
	// values which correspond with meters on the field. So B means the offside
	// line lies at 27.0.
	int iOffside = (int)(WM->getOffsideX( false) - 1.0 );
	if (iOffside < 26) // 0..25
		sprintf(strTmp, "%c", 'a' + max( 0, iOffside));
	else
		// 26...
		sprintf(strTmp, "%c", 'A' + min(iOffside - 26, 25));
	strcat(strMsg, strTmp);

	// First add al values to a positive interval, since then we don't need
	// another byte for the minus sign. then take one digit at a time
	double x= max(0,min( rint( posBall.getX() + 48.0), 99.0));
	sprintf(strTmp, "%c%c%c%c%c%c%c%c", '0' + ((int)(x ) % 100 ) / 10, '0'
			+ ((int)(x ) % 100 ) % 10, '0'
			+ ((int)(rint(posBall.getY() + 34.0)) % 100 ) / 10, '0'
			+ ((int)(rint(posBall.getY() + 34.0)) % 100 ) % 10, '0'
			+ ((int)((velBall.getX() + 2.7) * 10 )) / 10, '0'
			+ ((int)((velBall.getX() + 2.7) * 10 )) % 10, '0'
			+ ((int)((velBall.getY() + 2.7) * 10 )) / 10, '0'
			+ ((int)((velBall.getY() + 2.7) * 10 )) % 10);
	strcat(strMsg, strTmp);
	Log.log( 6560, "say (%d) %s\n", WM->getPlayerNumber() , strMsg);

	return;
}

/*! This method is called when a penalty kick has to be taken (for both the 
 goalkeeper as the player who has to take the penalty. */
void Player::performPenalty() {
	VecPosition pos;
	int iSide = (WM->getSide() == WM->getSidePenalty() ) ? -1 : 1;
	VecPosition posPenalty(iSide*(52.5 - SS->getPenDistX()), 0.0);
	VecPosition posAgent = WM->getAgentGlobalPosition();
	AngDeg angBody = WM->getAgentGlobalBodyAngle();

	SoccerCommand soc(CMD_ILLEGAL);
	static PlayModeT pmPrev = PM_ILLEGAL;

	// raise number of penalties by one when a penalty is taken
	if ( (WM->getSide() == SIDE_LEFT && pmPrev != PM_PENALTY_SETUP_LEFT
			&& WM->getPlayMode() == PM_PENALTY_SETUP_LEFT ) || (WM->getSide()
			== SIDE_RIGHT && pmPrev != PM_PENALTY_SETUP_RIGHT
			&& WM->getPlayMode() == PM_PENALTY_SETUP_RIGHT ))
		m_iPenaltyNr++;

	// start with player 11 and go downwards with each new penalty
	// if we take penalty 
	if (WM->isPenaltyUs() && WM->getPlayerNumber()
			== (11 - (m_iPenaltyNr % 11))) {
		if (WM->getPlayMode() == PM_PENALTY_SETUP_LEFT || WM->getPlayMode()
				== PM_PENALTY_SETUP_RIGHT) {
			pos = posPenalty - VecPosition(iSide*2.0, 0);
			if (fabs(posAgent.getX() ) > fabs(pos.getX() ) )
				pos = posPenalty;
			if (pos.getDistanceTo(posAgent) < 0.6) {
				pos = posPenalty;
				if (fabs(VecPosition::normalizeAngle((pos-posAgent).getDirection() - angBody) ) > 20)
					soc = turnBodyToPoint(pos);
			}
			//  pos = WM->getAgentGlobalPosition();
		} else if ( (WM->getPlayMode() == PM_PENALTY_READY_LEFT
				|| WM->getPlayMode() == PM_PENALTY_READY_RIGHT
				|| WM->getPlayMode() == PM_PENALTY_TAKEN_LEFT
				|| WM->getPlayMode() == PM_PENALTY_TAKEN_RIGHT )
				&& WM->isBallKickable() ) {
			soc = kickTo(
					VecPosition(iSide*52.5, ((drand48()<0.5) ? 1 : -1)*5.9),
					2.7);
		} else
			pos = posPenalty;
	} else if (formations->getPlayerType() == PT_GOALKEEPER) {
		if (WM->getAgentViewAngle() != VA_NARROW)
			ACT->putCommandInQueue(SoccerCommand(CMD_CHANGEVIEW, VA_NARROW,
					VQ_HIGH));

		// is penalty them, stop it, otherwise go to outside field
		pos = posPenalty;
		if (WM->isPenaltyThem() ) {
			pos = VecPosition(iSide*(52.5 - 2.0), 0.0);
			if (SS->getPenAllowMultKicks() == false) {
				if (WM->getPlayMode() == PM_PENALTY_TAKEN_LEFT
						|| WM->getPlayMode() == PM_PENALTY_TAKEN_RIGHT) {
					if (WM->isBallCatchable() )
						soc = catchBall();
					else
						soc = intercept( true);
				}
			} else if (pos.getDistanceTo(posAgent) < 1.0)
				soc = turnBodyToPoint(VecPosition( 0, 0) ) ;
			else
				soc = moveToPos(pos, 25);
		} else
			pos.setVecPosition(iSide * ( PITCH_LENGTH/2.0 + 2 ), 25);
	} else {
		pos = VecPosition( 5.0, VecPosition::normalizeAngle(iSide*(50 + 20
				*WM->getPlayerNumber())), POLAR);
	}

	if (soc.isIllegal() && WM->getAgentGlobalPosition().getDistanceTo(pos) < 0.8) {
		soc = turnBodyToPoint(posPenalty);
	} else if (soc.isIllegal() ) {
		soc = moveToPos(pos, 10);
	}
	if (WM->getAgentStamina().getStamina() < SS->getRecoverDecThr()*SS->getStaminaMax() + 500
			&& soc.commandType == CMD_DASH)
		soc.dPower = 0;

	ACT->putCommandInQueue(soc);
	ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, soc) );

	pmPrev = WM->getPlayMode();
}

/** GOS - Gets the index of possition in RoleCG where agent's role is stored.
 *
 * @param index Player's index whose ObjectT will be returned
 * @return Object type fo the player.
 */
ObjectT Player::getPlayerTypeRole(int index) {
	ObjectT o;
	switch (index) {

	// GOALIE agent
	case 0:
		o = OBJECT_TEAMMATE_1;
		break;

		// moji TEAMMATES
	case 1:
		o = OBJECT_TEAMMATE_2;
		break;
	case 2:
		o = OBJECT_TEAMMATE_3;
		break;
	case 3:
		o = OBJECT_TEAMMATE_4;
		break;
	case 4:
		o = OBJECT_TEAMMATE_5;
		break;
	case 5:
		o = OBJECT_TEAMMATE_6;
		break;
	case 6:
		o = OBJECT_TEAMMATE_7;
		break;
	case 7:
		o = OBJECT_TEAMMATE_8;
		break;
	case 8:
		o = OBJECT_TEAMMATE_9;
		break;
	case 9:
		o = OBJECT_TEAMMATE_10;
		break;
	case 10:
		o = OBJECT_TEAMMATE_11;
		break;
	default:
		o = OBJECT_TEAMMATE_1;
	}
	return o;
}

/** GOS - Gets the index of possition in RoleCG where agent's role is stored.
 *
 * @param Player whose index will be returned
 * @return index to RoleCG.<br />-1 is returned if <b>o</b> is not a teamate.
 */
int Player::getPlayerIndexRole(ObjectT o) {
	int index;
	switch (o) {

	// GOALIE agent
	case OBJECT_TEAMMATE_1:
		index = 0;
		break;

		// moji TEAMMATES
	case OBJECT_TEAMMATE_2:
		index = 1;
		break;
	case OBJECT_TEAMMATE_3:
		index = 2;
		break;
	case OBJECT_TEAMMATE_4:
		index = 3;
		break;
	case OBJECT_TEAMMATE_5:
		index = 4;
		break;
	case OBJECT_TEAMMATE_6:
		index = 5;
		break;
	case OBJECT_TEAMMATE_7:
		index = 6;
		break;
	case OBJECT_TEAMMATE_8:
		index = 7;
		break;
	case OBJECT_TEAMMATE_9:
		index = 8;
		break;
	case OBJECT_TEAMMATE_10:
		index = 9;
		break;
	case OBJECT_TEAMMATE_11:
		index = 10;
		break;
	default:
		index = -1;
	}
	return index;
}

/** GOS - Function computes and returns an array of roles.
 * @returns roles for each agent in the array.
 */
RoleCG * Player::getPlayerRoles() {

	// potencialy pre aktivneho hraca
	double activePotencial[11];

	// potencialy pre rolu receiver
	double receiverPotencial[11];

	// isCloseToBall: pre kazdeho hraca su 3 mozne stavy
	// 1  - nachadza sa v okruhu 6 metrov od lopty, je pri nej tak blizko ze mozeano, je tak blizko pri lopte, ze moze do nej kopnut
	// 0  - nachadza sa v okruhu 6 mestrov od lopty, ale nemoze do nej kopnut (loptu ma napr. super a pod...)
	// -1 - nachadza sa mimo okruhu 6 metrov

	int isCloseToBall[11];

	// role jednotlivym hracom, na 11-tej pozicii je moja rola
	RoleCG * roles = new RoleCG[11];

	int iIndex;
	int index;

	double vR = SS->getPlayerSpeedMax();
	double t;

	// na zaciatku su potencialy pre role rovne 0.0
	for (int i=0; i < 11; i++) {
		activePotencial[i] = 0.0;
		receiverPotencial[i] = 0.0;
		isCloseToBall[i] = -1;
		roles[i] = NO_ROLE;
	}

	// pridelime rolu GOALIE
	roles[0] = GOALIE;

	// zistime si poziciu lopty a jej rychlost	
	VecPosition b0 = WM->getBallPos();
	VecPosition v0 = WM->getGlobalVelocity(OBJECT_BALL);

	/*
	 // ak som nevidel loptu
	 if(v0.getX() == UnknownDoubleValue)	{
	 // TODO som passive
	 }
	 */

	// prejdeme cez mnozinu mojich spoluhracov
	ObjectSetT set = OBJECT_SET_TEAMMATES_NO_GOALIE;

	// a iterujeme ...
	for (ObjectT o = WM->iterateObjectStart(iIndex, set); o != OBJECT_ILLEGAL; o
			= WM->iterateObjectNext(iIndex, set)) {
		index = getPlayerIndexRole(o);

		// ak nastane nejaka chyba alebo pre brankara potencial nepocitame
		if (index != 0) {
			// zistime si poziciu hraca na ihrisku
			VecPosition r0 = WM->getGlobalPosition(o);

			// vypocitaj najmensi cas zachytenia lopty	    		    		    	
			t = ArsToolbox::computeLeastInterceptTime(b0, v0, r0, vR, 2.0);

			// vypocitaj aky ma hrac potencial vzhladom k role ACTIVE
			activePotencial[index] = 1.0 / t;

			// musime urcit, ci bude hrac PASSER alebo INTERCEPTOR
			double r0ToBallDistance = (r0 - b0).getMagnitude();

			// ak agent je tak blizko ze moze kopnut do lopty priamo
			if ((r0ToBallDistance < SS->getMaximalKickDist())) {
				isCloseToBall[index] = 1;
			}
			// teraz sa pozrieme ci je od nej vzdialeny aspon do 10 metrov ak do nej nemoze priamo kopnut
			else if (r0ToBallDistance < 10) {
				isCloseToBall[index] = 0;
			} else {
				isCloseToBall[index] = -1;
			}
		}
	}
	WM->iterateObjectDone(iIndex);

	// spoluhraci su urceni, este to iste urcim pre seba
	VecPosition r0 = WM->getAgentGlobalPosition();

	// nasledovne dva riadky netreba, lebo index 11 neexistuje a informacie o mne mam ulozene tiez v tom poli  		    		    	
	//t = ArsToolbox::computeLeastInterceptTime(b0, v0, r0, vR, 2.0);
	//activePotencial[11] = 1.0 / t;

	double r0ToBallDistance = (r0 - b0).getMagnitude();

	if ((r0ToBallDistance < SS->getMaximalKickDist())) {
		isCloseToBall[index] = 1;
	} else if (r0ToBallDistance < 10) {
		isCloseToBall[index] = 0;
	} else {
		isCloseToBall[index] = -1;
	}

	// potencialy su vyratane, konecne urcime hraca s rolou ACTIVE

	// urcime hraca s rolou ACTIVE
	double maxValue = activePotencial[0];
	int maxI = 0;

	// zacneme od 1, pretoze brankar (index == 0) nemoze byt ACTIVE	
	for (int i=1; i < 11; i++) {
		if (activePotencial[i] > maxValue) {
			maxValue = activePotencial[i];
			maxI = i;
		}
	}

	// nikto by nemal vacsi potencial ako brankar?
	// to by nemalo nastat, ale ajtak to osetrime
	if (maxI == 0) {
		for (int i=1; i < 11; i++) {
			roles[i] = PASSIVE;
		}

		return roles;
	} else {
		if (isCloseToBall[maxI] == 1) {
			roles[maxI] = ACTIVE_PASSER;
		} else if (isCloseToBall[maxI] == 0) {
			roles[maxI] = ACTIVE_INTERCEPTOR;
		} else {
			// ak nemoze byt ziadny active, nemoze byt ani ziadny interceptor
			for (int i=1; i < 11; i++)
				roles[i] = PASSIVE;

			return roles;
		}
	}

	// mame pridelene ROLU ACTIVE, teraz treba pridelit role RECEIVER,
	// ak sa teda daju pridelit

	// pridelime role RECEIVER
	for (ObjectT o = WM->iterateObjectStart(iIndex, set); o != OBJECT_ILLEGAL; o
			= WM->iterateObjectNext(iIndex, set)) {
		index = getPlayerIndexRole(o);

		if (index != 0) {

			// ak spoluhrac este nema pridelenu rolu
			if (roles[index] == NO_ROLE) {

				// zistime si poziciu hraca na ihrisku
				VecPosition r0 = WM->getGlobalPosition(o);

				// d_ib
				double r0ToBallDistance = (r0 - b0).getMagnitude();

				// d_ig
				double r0ToOpponentGoalDistance = (r0 - WM->getGlobalPosition(OBJECT_GOAL_R)).getMagnitude();

				if (r0ToBallDistance <= 28)
					receiverPotencial[index] = ((1 / this->getMax(1,
							r0ToOpponentGoalDistance)) + 1);
				else
					receiverPotencial[index] = (1 / this->getMax(1,
							r0ToOpponentGoalDistance));
			}
		}
	}

	// nasledovne riadky podla mna netreba, lebo rola pre mna je uz vypocitana v tom poli...
	// este pre seba urcim potencial k role RECEIVER ak nemam este ziadnu rolu
	/*if(roles[11] == NO_ROLE)
	 {
	 // zistime si poziciu hraca na ihrisku
	 VecPosition r0 = WM->getAgentGlobalPosition();
	 
	 // d_ib
	 double r0ToBallDistance = (r0 - b0).getMagnitude();

	 // d_ig
	 double r0ToOpponentGoalDistance = (r0 - WM->getGlobalPosition(OBJECT_GOAL_R)).getMagnitude();
	 
	 if(r0ToBallDistance <= 28)
	 receiverPotencial[11] = ((1 / this->getMax(1, r0ToOpponentGoalDistance)) + 1);
	 else
	 receiverPotencial[11] = (1 / this->getMax(1, r0ToOpponentGoalDistance));
	 }*/

	// ok, potencialy su priradene, priradime role RECEIVER
	int maxIReceiver;

	for (int j=0; j < 2; j++) {
		double maxValueReceiver = 0.0;

		for (int i=1; i < 11; i++) {
			if (roles[i] == NO_ROLE) {
				if (receiverPotencial[i] > maxValueReceiver) {
					maxIReceiver = i;
					maxValueReceiver = receiverPotencial[i];
				}
			}
		}

		roles[maxIReceiver] == RECEIVER;
	}

	return roles;
}

/** GOS - Function returns the role of the player.
 * 
 * @returns the role of mine in the current formation on the field
 */
//JH Nepouzivat! Neuvolnuje pole rol.
RoleCG Player::getMyRole() {
	RoleCG * allPlayerRoles = this->getPlayerRoles();
	int index = getPlayerIndexRole(WM->getAgentObjectType());

	return allPlayerRoles[index];
}

double Player::getMax(double val1, double val2) {
	if (val1 > val2)
		return val1;
	else
		return val2;
}

/** GOS - Transforms an angle specified by DirCG to double value.
 * 
 * @param globalDirection Global direction angle 
 * @returns global Direction in radians
 */
double Player::transformGlobalDirectionToAngleCG(DirCG globalDirection) {
	switch (globalDirection) {
	case N:
		return WM->getGlobalAngle(OBJECT_GOAL_R);
		break;
	case NE:
		return WM->getGlobalAngle(OBJECT_FLAG_R_B);
		break;
	case E:
		return WM->getGlobalAngle(OBJECT_FLAG_C_B);
		break;
	case SE:
		return WM->getGlobalAngle(OBJECT_FLAG_L_B);
		break;
	case S:
		return WM->getGlobalAngle(OBJECT_GOAL_L);
		break;
	case SW:
		return WM->getGlobalAngle(OBJECT_FLAG_L_T);
		break;
	case W:
		return WM->getGlobalAngle(OBJECT_FLAG_C_T);
		break;
	case NW:
		return WM->getGlobalAngle(OBJECT_FLAG_R_T);
		break;
	default:
		return 0.0;
	}

	return 0.0;
}

/** GOS - Value rule no. 1 of an interceptor player.
 * 
 * @param agentAction Chosen agent action 
 * @returns value
 */
double Player::valueRuleInterc1(PlayerActionCG agentAction) {
	if (agentAction.getActionType() == INTERCEPT)
		return 10.0; else
                return 0.0; //JH
}

/** GOS - Value rule no. 1 of a passer player.
 * 
 * @param playerRoles array of roles for every player on the field
 * @param teammateIndex index of a teammate of mine
 * @param agentAction chosen action of the current player
 * @param teammateAction chosen action of a teammate of mine
 * @param passDirection direction of a leading pass to a teammate of mine  
 * @returns value
 */
double Player::valueRulePasser1(RoleCG * playerRoles, int teammateIndex,
		PlayerActionCG agentAction, PlayerActionCG teammateAction,
		DirCG passDirection) {
	// if an incorrect input has been specified then the value of this rule is 0.0
	// <1-10> is allowed
	if (teammateIndex < 1 || teammateIndex > 10)return 0.0;

	// conditions of the value rule
	if(
			playerRoles[teammateIndex] == RECEIVER &&

			// TODO: neni este implementovana
			//(WM->isPassBlockedCG(11, teammateIndex, passDirection) == false) &&
			(isPassBlockedCG(getPlayerIndexRole(WM->getAgentObjectType()), teammateIndex, passDirection) == false) &&

			(agentAction.getActionType() == LEADING_PASS_TO &&
					agentAction.getTeammateIndex() == teammateIndex &&
					agentAction.getDirection() == passDirection) &&

			(teammateAction.getActionType() == MOVE_TO &&
					teammateAction.getDirection() == passDirection)
	)

	return uCG(teammateIndex, passDirection); else
        return 0.0; //JH
}

/** GOS - Value rule no. 2 of a passer player.
 * 
 * @param agentAction chosen action of the current player
 * @param emptySpaceDirection possible direction to an empty area (30m triangle)  
 * @returns value
 */
double Player::valueRulePasser2(PlayerActionCG agentAction,
		DirCG emptySpaceDirection) {
	if (isEmptySpace(getPlayerIndexRole(WM->getAgentObjectType()),
			emptySpaceDirection) && (agentAction.getActionType() == DRIBBLE
			&& agentAction.getDirection() == emptySpaceDirection)

	)
		return 2.0; else
                return 0.0; //JH
}

/** GOS - Value rule no. 3 of a passer player.
 * 
 * @param agentAction chosen action of the current player
 * @returns value
 */
double Player::valueRulePasser3(PlayerActionCG agentAction) {
	if (agentAction.getActionType() == CLEAR_BALL ///TODO NASA UPRAVA

	)
		return 0.1; else
                return 0.0;
}

/** GOS - Determines whether an object is in front of opponent's goal
 * 
 * @param pos global position of an object on the field
 * @returns value
 */
bool Player::isInFrontOfGoalCG(VecPosition pos) {
	ObjectT objFlag = (WM->getSide() == SIDE_LEFT ) ? OBJECT_FLAG_P_R_C
			: OBJECT_FLAG_P_L_C;

	VecPosition posFlag = SoccerTypes::getGlobalPositionFlag(objFlag,
			WM->getSide());

	if (pos.getX() > posFlag.getX() && fabs(pos.getY())
			< PENALTY_AREA_WIDTH/4.0)
		return true;

	return false;
}

/** GOS - Value rule no. 4 of a passer player.
 * 
 * @param agentAction chosen action of the current player
 * @returns value
 */
double Player::valueRulePasser4(PlayerActionCG agentAction) {
	if (isInFrontOfGoalCG(WM->getAgentGlobalPosition()) &&

	WM->isBallKickable() &&

	(agentAction.getActionType() == SCORE))
		return 10.0; else
                return 0.0; //JH
}

/** GOS - Value rule no. 1 of a receiver player.
 * 
 * @param playerRoles array of roles for every player on the field
 * @param teammateIndex index of a teammate of mine
 * @param agentAction chosen action of the current player
 * @param teammateAction chosen action of a teammate of mine
 * @param passDirection direction of a leading pass to a teammate of mine  
 * @returns value
 */
double Player::valueRuleReceiver1(RoleCG * playerRoles, int teammateIndex,
		PlayerActionCG agentAction, PlayerActionCG teammateAction,
		DirCG passDirection) {
	if (playerRoles[teammateIndex] == ACTIVE_INTERCEPTOR &&

	// TODO: neni este implementovana
			//(WM->isPassBlockedCG(11, teammateIndex, passDirection) == false) &&
                        (isPassBlockedCG(getPlayerIndexRole(WM->getAgentObjectType()), teammateIndex, passDirection) == false) && //JH

			(teammateAction.getActionType() == INTERCEPT) &&

	(agentAction.getActionType() == MOVE_TO && agentAction.getDirection()
			== passDirection))
		return uCG(getPlayerIndexRole(WM->getAgentObjectType()), passDirection); else
                return 0.0; //JH
}

/** GOS - Value rule no. 2 of a receiver player.
 * 
 * @param playerRoles array of roles for every player on the field
 * @param teammateIndex1 index of a teammate1 of mine
 * @param teammateIndex2 index of a teammate2 of mine
 * @param agentAction chosen action of the current player
 * @param teammateAction1 chosen action of a teammate1 of mine
 * @param passDirection1 pass direction from the current player to teammate1
 * @param teammateAction2 chosen action of a teammate2 of mine
 * @param passDirection2 pass direction from teammate1 to teammate2
 * @returns value
 */
double Player::valueRuleReceiver2(RoleCG * playerRoles, int teammateIndex1,
		int teammateIndex2, PlayerActionCG agentAction,
		PlayerActionCG teammateAction1, DirCG passDirection1,
		PlayerActionCG teammateAction2, DirCG passDirection2) {
	if (playerRoles[teammateIndex2] == RECEIVER &&

	// TODO: neni este implementovana
			//(WM->isPassBlockedCG(teammateIndex2, 11, passDirection1) == false) &&
                        (isPassBlockedCG(teammateIndex2, getPlayerIndexRole(WM->getAgentObjectType()), passDirection2) == false) && //JH

			(teammateAction1.getActionType() == LEADING_PASS_TO
					&& teammateAction1.getTeammateIndex() == teammateIndex2
					&& teammateAction1.getDirection() == passDirection2) &&

	(teammateAction2.getActionType() == MOVE_TO
			&& teammateAction2.getDirection() == passDirection2) &&

	(agentAction.getActionType() == MOVE_TO && agentAction.getDirection()
			== passDirection1))
		return uCG(getPlayerIndexRole(WM->getAgentObjectType()), passDirection1); else
                return 0.0; //JH
}

/** GOS - Value rule no. 3 of a receiver player.
 * 
 * @param agentAction chosen action of the current player
 * @returns value
 */
double Player::valueRuleReceiver3(PlayerActionCG agentAction) {
	if (agentAction.getActionType() == MOVE_TO_STRAT_POS)
		return 1.0; else
                return 0.0; //JH
}

/** GOS - Value rule no. 1 of a passive player.
 * 
 * @param agentAction chosen action of the current player
 * @returns value
 */
double Player::valueRulePassive1(PlayerActionCG agentAction) {
	if (agentAction.getActionType() == MOVE_TO_STRAT_POS)
		return 1.0; else
                return 0.0; //JH
}

/** GOS - Functions returns an optimal common action computed by the variable elimination algorithm.
 * 
 * @returns optimal joint actions of all agents excluding the goalie.
 */
PlayerActionCG * Player::getOptimalCommonAction() {
	PlayerActionCG * optimalPlayerActions;

	// priradime hracom role
	RoleCG * playerRoles = this->getPlayerRoles();

	// najdeme optimalnu spolocnu akciu pre vsetkych agentov
	optimalPlayerActions = variableEliminationAlgorithmCG(playerRoles);

	delete playerRoles; //JH

	return optimalPlayerActions;
}

/** GOS - Functions returns an optimal common action computed by the variable elimination algorithm.
 * 
 * @returns optimal joint actions of all agents excluding the goalie.
 */
PlayerActionCG * Player::variableEliminationAlgorithmCG(RoleCG * playerRoles) {

    // vybrane akcie agentov v doprednom prechode - faza eliminacie agentov
    PlayerActionCG * playerToEliminateAction[12];

    for (int i = 0; i < 12; i++)
    {
	playerToEliminateAction[i] = new PlayerActionCG();
    }

    // vsetkych agentov postupne eliminujeme z grafu v poradi zlava doprava (okrem brankara)
    for (int playerToEliminateIndex = 1; playerToEliminateIndex < 12; playerToEliminateIndex++)
    {
	// zistime si aku ma agent rolu
	RoleCG agentToEliminateRole = playerRoles[playerToEliminateIndex];

	if (agentToEliminateRole == ACTIVE_PASSER || agentToEliminateRole == ACTIVE_INTERCEPTOR || agentToEliminateRole == RECEIVER)
	{
	    // musime definovat ake prislusne akcie bude mat agent urceny na eliminaciu s ohladom na jeho rolu
	    PlayerActionTypeCG possibleAgentToEliminateActions[POSSIBLE_PLAYER_ACTION_TYPES_CG_NO];
	    int possibleAgentToEliminateActionsNo;

	    switch (agentToEliminateRole)
	    {
		case ACTIVE_PASSER:
		    possibleAgentToEliminateActions[0] = LEADING_PASS_TO;
		    possibleAgentToEliminateActions[1] = DRIBBLE;
		    possibleAgentToEliminateActions[2] = CLEAR_BALL;
		    possibleAgentToEliminateActions[3] = SCORE;

		    possibleAgentToEliminateActionsNo = 4;
		    break;

		case ACTIVE_INTERCEPTOR:
		    possibleAgentToEliminateActions[0] = INTERCEPT;

		    possibleAgentToEliminateActionsNo = 1;
		    break;

		case RECEIVER:
		    possibleAgentToEliminateActions[0] = MOVE_TO;
		    possibleAgentToEliminateActions[1] = MOVE_TO_STRAT_POS;

		    possibleAgentToEliminateActionsNo = 2;
		    break;

		default:
		    possibleAgentToEliminateActionsNo = 0;
	    }

	    // pre kazdeho spoluhraca agenta urceneho na eliminaciu
	    for (int teammateIndex = playerToEliminateIndex+1; teammateIndex < 12; teammateIndex++)
	    {
    		// okrem agenta urceneho na eliminaciu
		if (teammateIndex != playerToEliminateIndex)
		{
		    // akcia agenta na eliminaciu z koordinacneho grafu
		    PlayerActionCG * playerToEliminateAction;

		    // musime definovat ake prislusne akcie bude mat agent urceny na eliminaciu s ohladom na jeho rolu
		    PlayerActionTypeCG possibleTeammateActions[POSSIBLE_PLAYER_ACTION_TYPES_CG_NO];
		    int possibleTeammateActionsNo;

		    // zistime si jeho rolu
		    RoleCG teammateRole = playerRoles[teammateIndex];

		    switch (teammateRole)
		    {
			case ACTIVE_PASSER:
			    possibleTeammateActions[0] = LEADING_PASS_TO;
			    possibleTeammateActions[1] = DRIBBLE;
			    possibleTeammateActions[2] = CLEAR_BALL;
			    possibleTeammateActions[3] = SCORE;

			    possibleTeammateActionsNo = 4;
			    break;

			case ACTIVE_INTERCEPTOR:
			    possibleTeammateActions[0] = INTERCEPT;

			    possibleTeammateActionsNo = 1;
			    break;

			case RECEIVER:
			    possibleTeammateActions[0] = MOVE_TO;
			    possibleTeammateActions[1] = MOVE_TO_STRAT_POS;

			    possibleTeammateActionsNo = 2;
			    break;

			default:
			    possibleTeammateActionsNo = 0;
		    }

		    // ak ma rolu, ktora nas zaujima a s ktorou my ako agent mozeme koordinovat nase akcie
		    if (teammateRole == ACTIVE_PASSER || teammateRole == ACTIVE_INTERCEPTOR || teammateRole == RECEIVER)
		    {
			// maximalny vynos, ktory hrac urceny na eliminaciu moze dosiahnut
			double maxValue = 0.0;

			// skusime vsetky mozne kombinacie mojich akcii vzhladom k zvolenym akciam mojich spoluhracov
			for (int actionIndex=0; actionIndex < possibleAgentToEliminateActionsNo; actionIndex++)
			{
			    if (possibleAgentToEliminateActions[actionIndex] == LEADING_PASS_TO)
			    {
				playerToEliminateAction = new PlayerActionCG();
				playerToEliminateAction->setActionType(LEADING_PASS_TO);

				DirCG dir[8] = {N, NW, W, SW, S, SE, E, NE};
				int possibleDirectionsNo = 8;

				for (int directionIndex = 0; directionIndex < possibleDirectionsNo; directionIndex++)
				{

				    // v tomto okamihu uz mame zlozenu celu akciu agenta urceneho na eliminaciu
				    // teraz zlozime akcie jeho 

				    for (int teammateActionIndex=0; teammateActionIndex < 0; )
				    {
				    }

				}

			    }

			    double rulesCombinationValue = 0.0;

			    //rulesCombinationValue += valueRulePasser1(playerRoles, teammateIndex, agentAction, teammateAction, passDirection);
			}
		    }
		}
	    }

	    break;
	}
    }

    delete playerToEliminateAction[12];

    return NULL;
}

/** Implemented by Jahodovi Princovia, 2007/2008
 * 
 * @param playerIndex index of the player in role array 
 * @param teammateIndex index of teammate in role array
 * @param passDirection direction of the leading pass to the teammate specified by teammateIndex
 * 
 * return true / false
 */
bool Player::isPassBlockedCG(int playerIndex, int teammateIndex,
		DirCG passDirection) {

	double dConfThr = 0.5;

	ObjectT player = getPlayerTypeRole(playerIndex);
	ObjectT teammate = getPlayerTypeRole(teammateIndex);
	VecPosition playerVecPosition = WM->getGlobalPosition(player);
	VecPosition teammateVecPosition = WM->getGlobalPosition(teammate);

	double distTeammate = playerVecPosition.getDistanceTo(teammateVecPosition);

	// create line at this angle in 
	AngDeg angle = transformGlobalDirectionToAngleCG(passDirection);
	double x, y;
	ArsToolbox::createLineTo(playerVecPosition.getX(),
			playerVecPosition.getY(), angle, 1, &x, &y);

	int iIndex;
	for (ObjectT o = WM->iterateObjectStart(iIndex, OBJECT_SET_OPPONENTS,
			dConfThr); o != OBJECT_ILLEGAL; o = WM->iterateObjectNext(iIndex,
			OBJECT_SET_OPPONENTS, dConfThr) ) {
		//vector position for opponent player
		VecPosition opponentVecPosition = WM->getGlobalPosition(o);

		// get player position and his distance
		double dDist = playerVecPosition.getDistanceTo(opponentVecPosition);

		//if opponent player is beyond our player, then overstep
		if (dDist > distTeammate)
			continue;

		// find interest radius of circle for this player
		double r = getInterestRadius(dDist);

		double fake;
		if ((ArsToolbox::circleLineIntersection(playerVecPosition.getX(),
				playerVecPosition.getY(), x, y, opponentVecPosition.getX(),
				opponentVecPosition.getY(), r, &fake, &fake, &fake, &fake)

		) > 0)
			return false;
	}
	return true;
}

/** Implemented by Jahodovi Princovia, 2007/2008
 * 
 * @param playerIndex index of teammate in role array
 * @param direction of the possible player movement
 * 
 * return true / false
 */
bool Player::isEmptySpace(int playerIndex, DirCG direction) {

	VecPosition playerPos;  
	double coneLength;
	AngDeg playerAngle;
	
	ObjectT player = getPlayerTypeRole(playerIndex);
	
	playerPos = WM->getGlobalPosition(player);
	
	 if( playerPos.getX() > 0.0 )
		    coneLength = 6.0;
		  else
		    coneLength = 8.0;
	 
	 // create line at this angle in 
	 playerAngle = transformGlobalDirectionToAngleCG(direction);	 
	  
	 //overriden method isFreeCone( double dAngle, double dLength )
	 	 
	   VecPosition   posCircle;
	   Circle        c;
	   VecPosition   posEndCone;
	   VecPosition   posStartCone;
	 	
	   double x = cosDeg( playerAngle );
	   double y = sinDeg( playerAngle );
	   
	   posCircle.setX( playerPos.getX() + 1.5 * x );
	   posCircle.setY( playerPos.getY() + 1.5 * y );
	   c.setCircle( posCircle, 4.0 );

	   if( ( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, c ) != 0 ) )
	   	return false;
	   
	   posEndCone.setX( playerPos.getX() + x * coneLength );
	   posEndCone.setY( playerPos.getY() + y * coneLength );
	   
	   if( posEndCone.getX() > 52.5 || posEndCone.getX() < -52.5 )
	     return false;
	   if( posEndCone.getY() > 34.0 || posEndCone.getY() < -34.0 )
	     return false;
	   
	   posStartCone.setX( playerPos.getX() - x );
	   posStartCone.setY( playerPos.getY() - y );
	   
	   if( WM->getNrInSetInCone( OBJECT_SET_OPPONENTS, 2.5, 
	             posStartCone, posEndCone ) != 0 )
	     return false;
	   else
	     return true;   
}

/*Implemented by Jahodovi Princovia, 2007/2008
 * 
 * @param teammateIndex index of teammate in role array
 * @param passDirection direction of the pass
 * 
 * return A number from interval <5,7>. Now it returns a fake value 6.0.
 */
double Player::uCG(int teammateIndex, DirCG passDirection) {

	//TODO otestovat
	ObjectT teammate = getPlayerTypeRole(teammateIndex);
	VecPosition teammateVecPosition = WM->getGlobalPosition(teammate);

	double opponentDistanceSum = 0.0;
	double teammateDistanceSum = 0.0;
	double dConfThr = 0.9;
	int iIndex;	//indexovac

	for (ObjectT opponent = WM->iterateObjectStart(iIndex,
			OBJECT_SET_OPPONENTS, dConfThr); opponent != OBJECT_ILLEGAL; opponent
			= WM->iterateObjectNext(iIndex, OBJECT_SET_OPPONENTS, dConfThr) ) {

		//vector position for opponent player
		VecPosition opponentVecPosition = WM->getGlobalPosition(opponent);

		double dDist = teammateVecPosition.getDistanceTo(opponentVecPosition);

		//scitavanie jednotlivych vzdialenosti protihracov od prijimacieho hraca
		opponentDistanceSum += 1/dDist;

	}

	// find teammates  
	for (ObjectT teammate = WM->iterateObjectStart(iIndex,
			OBJECT_SET_TEAMMATES_NO_GOALIE, dConfThr); teammate
			!= OBJECT_ILLEGAL; teammate = WM->iterateObjectNext(iIndex,
			OBJECT_SET_TEAMMATES_NO_GOALIE, dConfThr) ) {

		//TODO otestovat - nestitavame vzdialenost od sameho seba
		if (teammateIndex == iIndex)
			continue;

		//vector position for opponent player
		VecPosition otherTeammateVecPosition = WM->getGlobalPosition(teammate);

		double dDist =
				teammateVecPosition.getDistanceTo(otherTeammateVecPosition);

		//scitavanie jednotlivych vzdialenosti protihracov od prijimacieho hraca
		teammateDistanceSum += 1/dDist;

	}

	//fitness prijimacieho hraca vzhladom na poziciu od superovej branky
	//vracia hodnoty od (-0.5, 0.5), kde 0.5 je najlepsie ohodnotenie fitness t.j. je
	//najblizsie k supervoej brane
	double positionFitness = teammateVecPosition.getX() / 100;

	double opponentDistanceMedian = (opponentDistanceSum / -11)
			+ (teammateDistanceSum / 11) + positionFitness;

	opponentDistanceMedian = opponentDistanceMedian / 4 + 6;

	//vracia hodnoty od 5,8 do 6,7
	return opponentDistanceMedian;

}
