/*LICENSE_BEGIN
 * ******************************************************************
 * All original parts of SoccerScope2003 have copyright (2003) by
 * Team YowAI, represented by Mr. Shuhei Shiota
 * The Department of Creative Informatics
 * The Graduate School of Information Science and Technology
 * The University of Tokyo
 * 
 * Additional classes (except libraries in /lib) were developed by 
 * the Dainamite Robocup-Team of the DAI-Labor, TU-Berlin 
 * (www.dai-labor.de)
 * 
 * Extended/new classes/packages are:
 *   soccerscope.analyzer.LogfileAnalyzer
 *   soccerscope.daistatistic.*
 *   soccerscope.daistatistic.charts.*
 *   soccerscope.daistatistic.tools.*
 *   soccerscope.db.*
 *   soccerscope.dialog.*
 *   soccerscope.model.PlayerWorldModel
 *   soccerscope.model.SimplePlayer
 *   soccerscope.net.AgentConnector
 *   soccerscope.training.*
 *   soccerscope.training.conditions.*
 * ******************************************************************
LICENSE_END*/
package soccerscope.analyzer;

import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedList;

import robocup.component.geometry.StraightLine;
import robocup.component.geometry.Vektor;
import robocup.component.worldobjects.Ball;
import robocup.component.worldobjects.Player;
import robocup.component.worldobjects.Player.RELATIVE;
import soccerscope.file.LogFileReader;
import soccerscope.model.ClientScene;
import soccerscope.model.HeteroParam;
import soccerscope.model.PlayMode;
import soccerscope.model.SceneBuilder;
import soccerscope.model.Team;
import soccerscope.util.geom.Point2f;
import soccerscope.util.geom.Vector2f;

public class LogfileAnalyzer {

	private static double MAX_DIST = 30;

	Ball ball = Ball.getDummyInstance();

	Player[] player = Player.getDummyInstances(22);

	public LogfileAnalyzer() {
		super();
	}

	/**
	 * This method analyzes a logfile.
	 */
	public LinkedList<double[]> analyzeLogfile(String filename) {
		LinkedList<ClientScene> list = createSceneListFromFile(filename);

		LinkedList<ClientScene> situations = new LinkedList<ClientScene>();

		// filter scenes, where a goal-shot can happen?
		for (int i = 0; i < list.size(); i++) {
			ClientScene scene = list.get(i);
			ClientScene nextScene = (i < list.size() - 1) ? list.get(i + 1)
					: null;
			if (hasPlayMode(scene, PlayMode.PM_PlayOn)
					&& playerCanShootGoal(scene)
					&& sceneContainsGoalShot(scene, nextScene)) {
				situations.add(scene);
			}
		}

		// System.out.println("\nSituations found: " + situations.size());

		boolean success[] = new boolean[situations.size()];
		Arrays.fill(success, false);

		for (int i = 0; i < situations.size(); i++) {
			boolean found = false;
			for (int j = 0; j < list.size(); j++) {
				if (!found) {
					if (situations.get(i) == list.get(j))
						found = true;
				} else {
					// stop after 30 cycles
					if (list.get(j).time > situations.get(i).time + 30)
						break;

					if (situations.size() > i + 1
							&& situations.get(i + 1).time <= list.get(j).time)
						break;

					if (list.get(j).pmode.pmode == PlayMode.PM_AfterGoal_Left
							|| list.get(j).pmode.pmode == PlayMode.PM_AfterGoal_Right) {
						success[i] = true;
						break;
					}
				}
			}
		}

		// store situations:
		LinkedList<double[]> result = new LinkedList<double[]>();
		for (int i = 0; i < situations.size(); i++) {
			ClientScene scene = situations.get(i);
			ClientScene next = null;
			for (int j = 0; j < list.size(); j++) {
				if (list.get(j) == scene)
					next = list.get(j + 1);
			}

			double[] traindata = createTrainingData(scene, next, success[i]);
			// System.out.println(Arrays.toString(traindata));
			result.add(traindata);
		}

		return result;
	}

	private double[] createTrainingData(ClientScene scene,
			ClientScene nextScene, boolean b) {
		int index = -1;
		for (int i = 0; i < scene.player.length; i++) {
			if (scene.player[i].isKickable(scene.ball.pos)) {
				index = i;
			}
			// init player
			player[i].setPosition(getVektor(scene.player[i].pos));
			player[i].setBodyDir(scene.player[i].angle);
			player[i].setGoalie(scene.player[i].isGoalie());
		}

		Vektor goalPos = new Vektor();
		if (index < 11)
			goalPos.y = 52.5;
		else
			goalPos.y = -52.5;

		// init ball
		ball.setPosition(getVektor(scene.ball.pos));
		ball.setSpeed(getVektor(scene.ball.vel));

		double[] data = new double[32];

		// 0 - angle of shooting player to goal
		// 1 - distance of shooting player to goal
		// 2 - angle of shooting player to ball
		// 3 - distance of shooting player to ball
		// 4 - kickable margin
		// 5 - relativer ballgeschwindigkeitswinkel
		// 6 - relative ballgeschwindigkeit
		// 7 - relativer schuwinkel
		// 8 - schussgeschwindigkeit
		// 9 - relativer winkel zum torwart
		// 10 - entfernung zum torwart
		// 11-30 : relativer winkel + entf zum nchsten gegenspieler, geordnet
		// ber winkel [-180, 180]
		data[0] = player[index].getAngleRel(goalPos, RELATIVE.TO_BODY) / 180.0;
		data[1] = player[index].getDistance(goalPos)
				/ Math.max(MAX_DIST, player[index].getDistance(goalPos));
		data[2] = player[index].getAngleRel(ball.getPosition(),
				RELATIVE.TO_BODY) / 180.0;
		data[3] = player[index].getDistance(ball)
				/ Math.max(MAX_DIST, player[index].getDistance(ball));
		data[4] = (HeteroParam.get(scene.player[index].type).kickable_margin) / 0.9;

		// prepare 5, 6
		Vektor nextBall = getVektor(nextScene.ball.pos);
		double nullAngle = nextBall.getAngleTo(goalPos);
		double spdAngle = getVektor(nextScene.ball.vel).getAngle();

		Vektor tmp = getVektor(scene.ball.vel);
		tmp.rotate(-nullAngle);
		data[5] = tmp.getAngle() / 180.0;
		data[6] = getVektor(scene.ball.vel).getLength()
				/ Math.max(MAX_DIST, getVektor(scene.ball.vel).getLength());
		data[7] = Vektor.normalize(spdAngle - nullAngle) / 180.0;
		data[8] = getVektor(nextScene.ball.vel).getLength() / 2.7;

		LinkedList<Player> playerOrdered = new LinkedList<Player>();
		Player goalie = null;
		for (int i = 0; i < player.length; i++) {
			if (index < 11 && i >= 11) {
				if (player[i].isGoalie())
					goalie = player[i];
				else
					playerOrdered.add(player[i]);
			} else if (index >= 11 && i < 11) {
				if (player[i].isGoalie())
					goalie = player[i];
				else
					playerOrdered.add(player[i]);
			}
		}

		data[9] = Vektor.normalize(getVektor(scene.ball.pos).getAngleTo(
				goalie.getPosition())
				- nullAngle) / 180.0;
		data[10] = getVektor(scene.ball.pos).getDistance(goalie.getPosition())
				/ Math.max(MAX_DIST, getVektor(scene.ball.pos).getDistance(
						goalie.getPosition()));

		for (int i = 0; i < playerOrdered.size(); i++) {
			playerOrdered.get(i).setBodyDir(
					Vektor.normalize(getVektor(scene.ball.pos).getAngleTo(
							playerOrdered.get(i).getPosition())
							- nullAngle));
		}

		int i = 0;
		while (playerOrdered.size() > 0) {
			double small = 180;
			Player p = null;
			int ind = -1;
			for (int j = 0; j < playerOrdered.size(); j++) {
				if (playerOrdered.get(j).getBodyDir() < small) {
					small = playerOrdered.get(j).getBodyDir();
					p = playerOrdered.get(j);
					ind = j;
				}
			}
			playerOrdered.remove(ind);
			data[11 + (2 * i)] = Vektor.normalize(getVektor(scene.ball.pos)
					.getAngleTo(p.getPosition())
					- nullAngle) / 180.0;
			data[11 + (2 * i) + 1] = getVektor(scene.ball.pos).getDistance(
					p.getPosition())
					/ Math.max(MAX_DIST, getVektor(scene.ball.pos).getDistance(
							p.getPosition()));
			i++;
		}

		data[31] = (b) ? 1 : 0;

		return data;
	}

	public Vektor getVektor(Point2f p) {
		return new Vektor(Vektor.XY, p.y, p.x);
	}

	public Vektor getVektor(Vector2f p) {
		return new Vektor(Vektor.XY, p.y, p.x);
	}

	/**
	 * 
	 * @param scene
	 * @param mode
	 * @return
	 */
	private boolean hasPlayMode(ClientScene scene, int mode) {
		return (scene.pmode.pmode == mode);
	}

	/**
	 * 
	 * @param scene
	 * @return
	 */
	private boolean playerCanShootGoal(ClientScene scene) {
		Ball ball = Ball.getDummyInstance();
		ball.setPosition(new Vektor(Vektor.XY, scene.ball.pos.y,
				scene.ball.pos.x));

		for (int i = 0; i < scene.player.length; i++) {
			if (!scene.player[i].isKickable(scene.ball.pos))
				continue;
			if (i < 11) {
				return (ball.getDistance(new Vektor(Vektor.XY, 0, 52.5)) <= MAX_DIST);
			} else {
				return (ball.getDistance(new Vektor(Vektor.XY, 0, -52.5)) <= MAX_DIST);
			}
		}
		return false;

	}

	/**
	 * 
	 * @param scene
	 * @return
	 */
	private boolean sceneContainsGoalShot(ClientScene scene,
			ClientScene nextScene) {
		Ball ball = Ball.getDummyInstance();
		ball.setPosition(new Vektor(Vektor.XY, scene.ball.pos.y,
				scene.ball.pos.x));

		int i = 0;
		for (i = 0; i < scene.player.length; i++) {
			if (scene.player[i].isKickable(scene.ball.pos))
				break;
		}

		Vektor nextBallPos = new Vektor(Vektor.XY, nextScene.ball.pos.y,
				nextScene.ball.pos.x);
		Vektor nextBallSpd = new Vektor(Vektor.XY, nextScene.ball.vel.y,
				nextScene.ball.vel.x);
		StraightLine ballLine = new StraightLine(nextBallPos, nextBallPos
				.add(nextBallSpd));

		StraightLine goalLine = null;
		if (i < 11)
			goalLine = new StraightLine(new Vektor(Vektor.XY, 34, 52.5),
					new Vektor(Vektor.XY, -34, 52.5));
		else
			goalLine = new StraightLine(new Vektor(Vektor.XY, 34, -52.5),
					new Vektor(Vektor.XY, -34, -52.5));

		Vektor intersect = goalLine.getIntersectionPoint(ballLine);
		if (intersect == null)
			return false;
		if (Math.abs(intersect.x) > 8) {
			return false;
		}

		// dribble!
		for (i = 0; i < nextScene.player.length; i++) {
			if (nextScene.player[i].isKickable(nextScene.ball.pos))
				return false;
		}

		if (Math
				.abs(nextBallSpd.getAngle() - nextBallPos.getAngleTo(intersect)) > 3)
			return false;

		if (nextBallSpd.getLength() < 1.5)
			return false;

		return true;
	}

	/**
	 * This method retrieves a list of all scenes of a given logfile.
	 * 
	 * @param logfile -
	 *            the log-files name
	 * @return - the list containing all scenes of a game!
	 */
	private LinkedList<ClientScene> createSceneListFromFile(String logfile) {
		LinkedList<ClientScene> list = new LinkedList<ClientScene>();
		try {
			byte[] packet = new byte[4096];
			LogFileReader builder = new LogFileReader(logfile);

			ClientScene clientScene = null;
			PlayMode playmode = null;
			Team left = null;
			Team right = null;

			while (true) {
				int size;
				try {
					size = builder.readPacket(packet);
					if (size == -1) {
						return list;
					}
				} catch (IOException ie) {
					System.err.println(ie);
					return list;
				}

				switch (builder.getPacketType(packet)) {
				case SceneBuilder.SHOW_MODE:
					// here, a client-scene is created during game,
					clientScene = builder.makeScene(packet, playmode, left,
							right);
					list.add(clientScene);
					break;

				case SceneBuilder.MSG_MODE:
					builder.addMessage(clientScene, new String(packet, 6,
							size - 7));
					break;

				case SceneBuilder.DRAW_MODE:
					break;

				case SceneBuilder.BLANK_MODE:
					break;

				case SceneBuilder.PM_MODE:
					playmode = builder.makePlayMode(packet);
					break;

				case SceneBuilder.TEAM_MODE:
					left = builder.makeLeftTeam(packet, 2);
					right = builder.makeRightTeam(packet, 2);
					break;

				case SceneBuilder.PT_MODE:
					builder.makeHeteroParam(packet);
					break;

				case SceneBuilder.PARAM_MODE:
					builder.printServerParameter(packet);
					break;

				case SceneBuilder.PPARAM_MODE:
					builder.printPlayerParameter(packet);
					break;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return list;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		LogfileAnalyzer lfa = new LogfileAnalyzer();
		LinkedList<double[]> ll = lfa.analyzeLogfile(args[0]);
		System.err.println("\nScenes found: " + ll.size());
	}

}
